"""
    Written By:         Philip Uren
    Date of Creation:   28th Jan 2010
    Version:            0.1
    Description:
                        Implementation of the expectation maximisation (EM) algorithm for locating motifs in
                        biological (or any kind of) sequences

                        Note:   this code was written for educational purposes only. It has not been thoroughly tested
                                and may (probably does) contain bugs. There are inefficiencies in the computations - it
                                was written for ease of understanding rather than speed.

    TODO:
                        *.) Currently does 10 iterations - should instead check for changes and stop iterating when
                            none are found 
                        *.) Program currently translates characters to integers for lookup - change to associative
                            approach?
                        *.) No error checking present to make sure all sequences provided are the same length
                        *.) No error checking present to make sure sequences contain only expected characters
                        *.) Sequences and motif length hard-coded in script
                        *.) Remove magic numbers associated with alphabet length of 4
                        *.) Standardise function header comments
"""

import numpy
PSEUDO_COUNT = 0.0001               # add this pseudo count to character counts to avoid zero-counts

class Model :
    """
        A simple structure to wrap up the PWM and the background models 
    """
    def __init__(self, pwm = None, background = None) :
        self.pwm = pwm
        self.background = background

    def __str__(self):
        return str(self.pwm) + "\n" + str(self.background)


def prob(seq, model, j, width) :
    """
        calculate the probability of the sequence 'seq' given the assumption
        that the motif in this sequence begins at the location j and using the
        current model.
        that is: P(X[i] | Z[i,j] = 1, p)
        where p is the model (pwm and background)
    """
    prod = 1
    for k in range(0,len(seq)) :
        if (k <= j - 1) or (k >= j + width) :
            prod *= model.background[alphabetLookup(seq[k])]
        else :
            prod *= model.pwm[alphabetLookup(seq[k])][k-j]
    return prod

def expectation(sequence, model, width) :
    """
        Do the expectation step of EM - calculate the probabilities of there
        being a motif at each position j within each sequence i given the
        current model
    """

    sum_prob = 0
    for i in range(0,len(sequence)) :
        for j in range(0, len(sequence[i])) :
            sum_prob += prob(sequence[i],model,j, width)

    z = numpy.zeros([len(sequence), len(sequence[0])], float)
    for i in range(0,len(sequence)) :
        for j in range(0, len(sequence[i])) :
            print "doing z calc -- prob is: " + str(prob(sequence[i],model,j, width))
            z[i][j] = prob(sequence[i],model,j, width) / sum_prob

    return z

def alphabetLookup(ch) :
    if (ch=='A') : return 0
    if (ch=='C') : return 1
    if (ch=='T') : return 2
    if (ch=='G') : return 3

def probabilisticCount(probabilities, sequences, width):
    """
        returns an 4 x n matrix, where n is the length of the sequences. 
        think of this as moving the motif window along each sequence
    """
    countMatrix = numpy.zeros([4,width], float)

    # add a small pseudo count for each position
    for r in countMatrix :
        for c in r :
            c += PSEUDO_COUNT
    
    for i in range(0,len(sequences)) :
        for j in range(0, len(sequences[i]) - width + 1) :
            pr = probabilities[i][j]
            for k in range(0,width):
                ch = alphabetLookup(sequences[i][j + k])
                countMatrix[ch][k] += pr

    return countMatrix

def backgroundProbabilisticCounts(probabilities, sequences):
    counts = numpy.zeros([4], float)

    # add a small pseudo count for each chr
    for r in counts :
        r += PSEUDO_COUNT
    
    for i in range(0,len(sequences)) :
        for j in range(0, len(sequences[i])) :
            pr = 1 - probabilities[i][j]
            counts[alphabetLookup(sequences[i][j])] += pr
    return counts

def maximisation(sequences, probabilities, width) :
    """
        The maximisation step of the EM algorithm - generate the PWM and background model based
        on the current probabilities 
    """

    sequenceLengths = len(sequences[0])
    counts = probabilisticCount(probabilities, sequences, width)
    bgCounts = backgroundProbabilisticCounts(probabilities, sequences)
    pwm = numpy.zeros([4,width], float)
    bg = numpy.zeros([4], float)

    # calculate sum of counts for all base pairs at each column - motif model
    sums = numpy.zeros([width], float)
    for c in range(0,4) :
        for k in range(0,width) :
            sums[k] += counts[c][k]

    # calculate the pwm entries for the motif model
    for c in range(0,4) :
        for k in range(0,width) :
            pwm[c][k] = counts[c][k] / float(sums[k])

    # calculate sum of counts for all base pairs - background model
    sums = 0
    for c in range(0,len(bgCounts)) :
        sums += bgCounts[c]

    # calculate entries for bg model
    for c in range(0,4) :
        bg[c] = bgCounts[c] / sums

    # wrap up model and return
    return Model(pwm, bg)



def expectationMaximisation(sequences, width) :
    # generate a random PWM and background model as starting point
    pwm = numpy.random.rand(4,width)
    bg = numpy.random.rand(4)
    model = Model(pwm, bg)

    print model
    print "----\n"
    
    for i in range(0,10) :
        zs = expectation(sequences, model, width)
        model = maximisation(sequences, zs, width)
        print "zs are:" + "\n" + str(zs) + "\n"
        print "model is: "
        print model
        print "----\n"


# entry point...
seqs = ["ACTGGT", "ACTACT"]
expectedMotifLength = 3
expectationMaximisation(seqs, expectedMotifLength)
